
/*****************************************************************************/
/* sersamp.cpp                                                               */
/* Demonstrates using the POSSerialDevice object to interact with a 4690     */
/* application that uses serial communications ports 1 and 2.   This         */
/* sample expectes to be communicating with a TC terminal session running    */
/* the sample CBASIC application 'sertest.286'.  sertest.286 source and      */
/* executable are included with this sample                                  */
/*                                                                           */
/* Author: John Foley on 10/30/2001                                          */
/*                                                                           */
/* Copyright 2001, QVS Software, Inc.                                        */
/* All rights reserved.                                                      */
/*                                                                           */
/*****************************************************************************/


                                       /* import the type libray, us no      */
                                       /* namespace so we don't have to      */
                                       /* preface everything with QVSRPAM::  */
#include "stdafx.h"
#include <windows.h>
#include <objbase.h>
#include "qvs_rcs.h"
#import "c:\\tc_nt\\bin\\qvsrpam.exe" no_namespace

                                       /* prototypes                         */
long GetFCodeFromKey(int InKey);
void hexlog(char far *buff, int len);

                                       /* globals                            */
HANDLE OnlineEvt;

/* local 'keymap' */
typedef struct keymap
   {
   int Inkey;
   long FCode;
   } KEYMAP;

KEYMAP KeyMap[] =
   {
      {'/', 78},     // /
      {'0', 48},     // 0
      {'1', 49},     // 1
      {'2', 50},     // 2
      {'3', 51},     // 3
      {'4', 52},     // 4
      {'5', 53},     // 5
      {'6', 54},     // 6
      {'7', 55},     // 7
      {'8', 56},     // 8
      {'9', 57},     // 9
      {'s', 61},     // Signon
      {'c', 73},     // Clear
      {'e', 80},     // Enter
      {'t', 81},     // Total
      {'h', 91},     // Cash
      {0  , 0 }      // End of data
   };

/*
class definition for an object that knows how to deal with
events fired by the outgoing QVSRPAMEvents interface of the
RPAMTerminal object
*/
interface QVSRPAMEvents : public _IQVSRPAMEvents
   {
   /*
   our reference count
   */
   long m_cref;

   /*
   pointer to the Terminal we're sinking events for 
   */
   IRPAMTerminalPtr pTerm;

   protected:

   /*
   protected constructor means this object can only be created on the heap with new
   */
   virtual ~QVSRPAMEvents();

   public:
   /*
   this object implements the standard IDispatch members
   */
   virtual HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void __RPC_FAR *__RPC_FAR *ppvObject);
   virtual ULONG STDMETHODCALLTYPE   AddRef(void);
   virtual ULONG STDMETHODCALLTYPE   Release(void);
   virtual HRESULT STDMETHODCALLTYPE GetTypeInfoCount(UINT __RPC_FAR *pctinfo);
   virtual HRESULT STDMETHODCALLTYPE GetTypeInfo(UINT iTInfo, LCID lcid, ITypeInfo __RPC_FAR *__RPC_FAR *ppTInfo);
   virtual HRESULT STDMETHODCALLTYPE GetIDsOfNames(REFIID riid, LPOLESTR __RPC_FAR *rgszNames, UINT cNames,
                                                   LCID lcid, DISPID __RPC_FAR *rgDispId);
   virtual HRESULT STDMETHODCALLTYPE Invoke(DISPID dispIdMember, REFIID riid, LCID lcid, WORD wFlags,
                                            DISPPARAMS __RPC_FAR *pDispParams, VARIANT __RPC_FAR *pVarResult,
                                            EXCEPINFO __RPC_FAR *pExcepInfo, UINT __RPC_FAR *puArgErr);

   QVSRPAMEvents(IRPAMTerminalPtr Term);

   /*
   routine to hook this object up to receive events
   from the RPAMTerminal object
   */
   HRESULT HookEvents(IUnknown *pUnkObject, DWORD Cookie);
   };


/*****************************************************************************/
/* Mainline code                                                             */
/*****************************************************************************/
int main(int argc, char* argv[])
{
   long rc;
   /*
   initalize COM, need to use CoInitializeEx because
   QVSRPAM is an out of proc server
   */
   printf("Serial port demo program\n");
   CoInitializeEx(0, COINIT_MULTITHREADED);

   
   /*
   Create an instance of an App object and
   a Terminal object.  Constructor will call
   CoCreateInstance(), AddRef(), etc.
   */
   IRPAMTerminalPtr pTerm(__uuidof(RPAMTerminal));
   IRPAMAppPtr pApp(__uuidof(RPAMApp));

   /*
   create an event handler object
   */
   QVSRPAMEvents *EvtHandler = new QVSRPAMEvents(pTerm);

   /*
   Turn QVSRPAM tracing on.
   */
   pApp->PutTracing(1);

   /*
   Tell the library that we want to assign
   terminal numbers
   */
   pApp->PutTermNumAssignMode(rpaAppTermNum);

   /*
   Hook our implementation of IDispatch up to
   the RPAMTerminal object.  He will use this
   interface to fire events back to us.
   */
   EvtHandler->HookEvents(pTerm, 0);

   /*
   Set the terminal's unit ID.  Can be used
   to reconnect to a persistent session at
   a later time.
   */
   pTerm->PutUnitID(1);

   /*
   Enable both serial ports
   */
   pTerm->GetSerial1()->put_DeviceEnabled(1);
   pTerm->GetSerial2 ()->put_DeviceEnabled(1);
   

   /*
   We're going to be terminal 50
   */
   pTerm->PutTermNum(50);

   /*
   Connect to TC!
   */
   pTerm->Connect();

   /*
   Now wait for "W008" to appear on the 2x20,
   this is how we know our terminal is online to
   the controller.  The Online event only tells
   us that we're online to TC...
   */
   OnlineEvt = CreateEvent(NULL, 0, 0, "");

   printf("Waiting for online to the controller\n");
   rc = WaitForSingleObject(OnlineEvt, 60000);
   if (rc != WAIT_OBJECT_0)
      {
      printf("Timed out waiting for online!\n");
      return 1;
      }

   printf("Online to controller");

   /*
   get a pointer to the IOProcessor of the terminal
   so we can send keystrokes
   */
   IIOProcessorPtr IOProc = pTerm->GetIOProcessor();


   /*
   Setup to read characters from the console
   */
   HANDLE hStdIn  = GetStdHandle(STD_INPUT_HANDLE);
   INPUT_RECORD InRec;
   DWORD nRec = 1;
   CHAR InpChar;
   long FCode;
   BOOL DispPrompt=1;

   FlushConsoleInputBuffer(hStdIn);

   while(1)
      {
      /*
      Display the prompt and wait for a key down message
      */
      if (DispPrompt)
         {
         printf("\nEnter a key to send to the sales app (ENTER to exit): ");
         }

      DispPrompt = 0;

      /*
      Read from the console
      */
      ReadConsoleInput(hStdIn, &InRec, 1, &nRec);

      /*
      Ignore anything but key downs for keys between '0' and 'z'
      */
      if (InRec.EventType == KEY_EVENT &&
          InRec.Event.KeyEvent.bKeyDown)
         {
         InpChar = InRec.Event.KeyEvent.uChar.AsciiChar;
         /*
         Enter gets us out of the loop
         */
         if (InpChar == 0x0D)
            {
            printf("\nexiting");
            break;
            }
         else if (InpChar >= 30 && InpChar <= 122)
            {
            printf("%c", InpChar);

            /*
            special keys to cause program to
            excersize various functionality
            of the QVSRPAM library
            */
            if (0)
               {
               }

            /*
            look up the keystroke in the key map
            and if found send it to the sales app.
            */
            else if (FCode=GetFCodeFromKey(InpChar))
               {
               IOProc->SendKeyboardData(FCode);
               }
            else
               {
               printf("\nUnmapped key, ignoring");
               }
            /*
            need to redisplay the prompt
            */
            DispPrompt = 1;
            }
         }
      }

   /*
   the pApp and pTerm _com_ptr_t objects will
   call Release on their interfaces when their
   destructors are called
   */
   return 0;
}



/*****************************************************************************/
/* Event handler methods                                                     */
/*****************************************************************************/
/*
The event handler class only responds to
queries for _IQVSRPAMEvents and IUnknown
*/
HRESULT STDMETHODCALLTYPE QVSRPAMEvents::QueryInterface(REFIID riid, void __RPC_FAR *__RPC_FAR *ppvObject)
   {
   HRESULT hr = E_NOINTERFACE;

   if (IsEqualIID(riid, __uuidof(_IQVSRPAMEvents)) ||
       IsEqualIID(riid, __uuidof(IUnknown)))
      {
      *ppvObject = this;
      AddRef();
      hr = S_OK;
      }

   return hr;
   }

/*
basic AddRef() implementation
*/
ULONG STDMETHODCALLTYPE QVSRPAMEvents::AddRef(void)
   {
   return InterlockedIncrement(&m_cref);
   }

/*
basic Release() implementation
*/
ULONG STDMETHODCALLTYPE QVSRPAMEvents::Release(void)
   {
   long res;
   res = InterlockedDecrement(&m_cref);
   if (res == 0)
      delete this;

   return res;
   }

/*
QVSRPAM will never call this method on this object
*/
HRESULT STDMETHODCALLTYPE QVSRPAMEvents::GetTypeInfoCount(UINT __RPC_FAR *pctinfo)
   {
   return E_NOTIMPL;
   }

/*
QVSRPAM will never call this method on this object
*/
HRESULT STDMETHODCALLTYPE QVSRPAMEvents::GetTypeInfo(UINT iTInfo, LCID lcid, ITypeInfo __RPC_FAR *__RPC_FAR *ppTInfo)
   {
   return E_NOTIMPL;
   }

/*
QVSRPAM will never call this method on this object
*/
HRESULT STDMETHODCALLTYPE QVSRPAMEvents::GetIDsOfNames(REFIID riid, LPOLESTR __RPC_FAR *rgszNames,
                                                       UINT cNames, LCID lcid, DISPID __RPC_FAR *rgDispId)
   {
   return E_NOTIMPL;
   }

/*
This routine is where the RPAMTerminal object will call when he has
an event to tell us about
*/
HRESULT STDMETHODCALLTYPE QVSRPAMEvents::Invoke(DISPID dispIdMember, REFIID riid, LCID lcid, WORD wFlags,
                                                DISPPARAMS __RPC_FAR *pDispParams, VARIANT __RPC_FAR *pVarResult,
                                                EXCEPINFO __RPC_FAR *pExcepInfo, UINT __RPC_FAR *puArgErr)
   {
   switch(dispIdMember)
      {
      /* terminal events **************************************/
      /* OnlineEvent(void) */
      case 001:
         break;
      /* OfflineEvent(void) */
      case 002:
         break;
      /* ConnectTimeoutEvent(void) */
      case 003:
         break;

      /* display events **************************************/
      /* DisplayEvent(long Device, BSTR Row1Text, BSTR Row2Text) */
      case 101:
         {
         long Device = pDispParams->rgvarg[2].lVal;
         BSTR Row1Text = pDispParams->rgvarg[1].bstrVal;
         BSTR Row2Text = pDispParams->rgvarg[0].bstrVal;
         char temp1[50], temp2[50];
         char *dptr;

         wcstombs(temp1, Row1Text, wcslen(Row1Text));
         temp1[wcslen(Row1Text)] = 0;
         wcstombs(temp2, Row2Text, wcslen(Row2Text));
         temp2[wcslen(Row2Text)] = 0;

         printf("DISPLAY DATA: [%s][%s]\n", temp1, temp2);
         
         if (Device == 0)
            {
            if (!wcsncmp(Row1Text, L"W008", 4))
               {

                                       /* wake up the main thread            */
               SetEvent(OnlineEvt);
               }
            else if (!wcsncmp(Row1Text, L"SEND PORT 1", 11))
               {
                                       /* appl ready to read serial 1        */
               SAFEARRAY *sa;
               sa = SafeArrayCreateVector(VT_UI1, 0, 25);
               SafeArrayAccessData(sa, (void **)&dptr);
               strcpy(dptr, "[SERDATA1SERDATA1SERDATA]");
               SafeArrayUnaccessData(sa);

               pTerm->GetSerial1()->Write(25, &sa);
               SafeArrayDestroy(sa);
               }
           else if (!wcsncmp(Row1Text, L"SEND PORT 2", 11))
               {
                                       /* appl ready to read serial 2       */
               SAFEARRAY *sa;
               sa = SafeArrayCreateVector(VT_UI1, 0, 22);
               SafeArrayAccessData(sa, (void **)&dptr);
               strcpy(dptr, "[SERDATA2SERDATA2SERA]");
               SafeArrayUnaccessData(sa);

               pTerm->GetSerial2()->Write(25, &sa);
               SafeArrayDestroy(sa);
               }
            }
         }
         break;
      /* DisplayEventEx(long Device, BSTR Row1Text, BSTR Row2Text, long UpdateFlags)  */
      case 102:
         {
         BSTR Row1Text = pDispParams->rgvarg[2].bstrVal;
         BSTR Row2Text = pDispParams->rgvarg[1].bstrVal;
         long UpdateFlags = pDispParams->rgvarg[0].lVal;
         }
         break;

      /* printer events **************************************/
      /* PrintEvent(long Station, BSTR PrintLine, long LineFeeds) */
      case 201:
         {
         long Station = pDispParams->rgvarg[2].lVal;
         BSTR PrintLine = pDispParams->rgvarg[1].bstrVal;
         long LineFeeds = pDispParams->rgvarg[0].lVal;
         }
         break;

      /* ioproc events **************************************/
      /* AlphaEntryModeEnterEvent(void) */
      case 301:
         break;
      /* AlphaEntryModeExitEvent(void)  */
      case 302:
         break;
      /* FieldLevelInputEvent(long MaxLen, long EchoInput, long FirstDisplayCol, long LastDisplayCol, long StateNum) */
      case 303:
         {
         long MaxLen = pDispParams->rgvarg[4].lVal;
         long EchoInput = pDispParams->rgvarg[3].lVal;
         long FirstDisplayCol = pDispParams->rgvarg[2].lVal;
         long LastDisplayCol = pDispParams->rgvarg[1].lVal;
         long StateNum = pDispParams->rgvarg[0].lVal;
         }
         break;
      /* LockEvent(void) */
      case 304:
         break;
      /* ReadEvent(void) */
      case 306:
         break;
      /* UnlockEvent(void) */
      case 307:
         break;

      /* term control events **************************************/
      /* ActivateMarqueeDisplayEvent(long StartCol, long EndCol, long Delay) */
      case 401:
         {
         long StartCol = pDispParams->rgvarg[2].lVal;
         long EndCol = pDispParams->rgvarg[1].lVal;
         long Delay = pDispParams->rgvarg[0].lVal;
         }
         break;
      /* DeactivateMarqueeDisplayEvent(void) */
      case 402:
         break;
      /* LoadMarqueeDisplayEvent(long Segment, BSTR Data, long Length) */
      case 403:
         {
         long Segment = pDispParams->rgvarg[2].lVal;
         BSTR Data  = pDispParams->rgvarg[1].bstrVal;
         long Length = pDispParams->rgvarg[0].lVal;
         }
         break;
      /* OfflineTotalsDataEvent(SAFEARRAY(unsigned char) *Data, long DataLen) */
      case 404:
         {
         SAFEARRAY *Data = *(pDispParams->rgvarg[1].pparray);
         long DataLen = pDispParams->rgvarg[0].lVal;
         }
         break;
      /* PrintEjectEvent(void) */
      case 405:
         break;
      /* ReceiptCutEvent(void) */
      case 406:
         break;
      /* TermIPLControlEvent(BOOL Enable) */
      case 407:
         {
         BOOL Enable = pDispParams->rgvarg[0].boolVal;
         }
         break;
      /* TermReloadEvent(void) */
      case 408:
         break;
      /* UserDataEvent(SAFEARRAY(unsigned char) *Data, long DataLen) */
      case 409:
         {
         SAFEARRAY *Data = *(pDispParams->rgvarg[1].pparray);
         long DataLen = pDispParams->rgvarg[0].lVal;
         }
         break;
      /* RunningTotalDataEvent(unsigned long RunningTotal) */
      case 412:
         {
         long DataLen = pDispParams->rgvarg[0].lVal;
         }
         break;

      /* scanner events **************************************/
      /* ScannerUnlockEvent(void) */
      case 410:
         break;
      /* ScannerLockEvent(void) */
      case 411:
         break;

      /* keyboard events **************************************/
      /* WaitLightEvent(BOOL NewValue) */
      case 501:
         {
         BOOL NewValue = pDispParams->rgvarg[0].boolVal;
         }
         break;
      /* MessageLightEvent(BOOL NewValue) */
      case 502:
         {
         BOOL NewValue = pDispParams->rgvarg[0].boolVal;
         }
         break;
      /* OfflineLightEvent(BOOL NewValue) */
      case 503:
         {
         BOOL NewValue = pDispParams->rgvarg[0].boolVal;
         }
         break;
      /* FieldLevelInputEvent2(long MaxLen, long EchoInput, long FirstDisplayCol, long LastDisplayCol, long StateNum, BOOL AlphaEntry, char TransitionFlag) */
      case 504:
         {
         long MaxLen = pDispParams->rgvarg[6].lVal;
         long EchoInput = pDispParams->rgvarg[5].lVal;
         long FirstDisplayCol = pDispParams->rgvarg[4].lVal;
         long LastDisplayCol = pDispParams->rgvarg[3].lVal;
         long StateNum = pDispParams->rgvarg[2].lVal;
         BOOL AlphaEntry = pDispParams->rgvarg[1].boolVal;
         char TransitionFlag = pDispParams->rgvarg[0].bVal;

         }
         break;

      /* SerialOpenEvent(long Port, long Speed, BSTR Parity, long DataBits, long StopBits, long BuffSize); */
      case 517:
         {
         long Port = pDispParams->rgvarg[5].lVal;
         long Speed = pDispParams->rgvarg[4].lVal;
         BSTR Parity = pDispParams->rgvarg[3].bstrVal;
         long DataBits = pDispParams->rgvarg[2].lVal;
         long StopBits = pDispParams->rgvarg[1].lVal;
         long BuffSize = pDispParams->rgvarg[0].lVal;
         char tmp[2];
         wcstombs(tmp, Parity, 1);
         tmp[1] = 0;
         printf("Open for serial port %d: %d/%s/%d/%d size %d\n", 
                           Port, Speed, tmp, DataBits, StopBits, BuffSize);

         if (Port == 1)
            {
            pTerm->GetSerial1()->SendOpenRc(0);
            }
         else if (Port == 2)
            {
            pTerm->GetSerial2()->SendOpenRc(0);
            }
         else
            {
            printf("Unhandled open for port %d\n", Port);
            }
         }

         break;

      /* SerialCloseEvent(long Port); */
      case 518:
         {
         long Port = pDispParams->rgvarg[0].lVal;
         printf("Close for serial port %d\n", Port);
         }
         break;

      /* SerialWriteEvent(long Port, long Length, SAFEARRAY(unsigned char) *Data); */ 
      case 519:
         {
         long Port = pDispParams->rgvarg[2].lVal;
         long Length = pDispParams->rgvarg[1].lVal;
         SAFEARRAY *sa = pDispParams->rgvarg[0].parray;
         char *dptr;
         SafeArrayAccessData(sa, (void **)&dptr);

         printf("Write for serial port %d, data:\n", Port);
         hexlog(dptr, Length);
         SafeArrayUnaccessData(sa);
         if (Port == 1)
            {
            pTerm->GetSerial1()->SendWriteRc(Length);
            }
         else if (Port ==2)
            {
            pTerm->GetSerial2()->SendWriteRc(Length);
            }
         else
            {
            printf("Unhandled write for port %d\n");
            }
         
         }
         break;
      /* SerialPutLongEvent(long Port, long Timeout, long PortStatusInfo); */
      case 520:
         {
         long Port = pDispParams->rgvarg[2].lVal;
         long TimeOut = pDispParams->rgvarg[1].lVal;
         long StatusValue  = pDispParams->rgvarg[0].lVal;
         printf("PutLong for serial port %d: %d %d\n", Port, TimeOut, StatusValue);
         if (Port == 1)
            {
            pTerm->GetSerial1()->SendPutLongRc(0);
            }
         else if (Port ==2)
            {
            pTerm->GetSerial2()->SendPutLongRc(0);
            }
         else
            {
            printf("Unhandled putlong for port %d\n");
            }
         }
         break;
      }
   return S_OK;
   }

/*
Event handler constructor
*/
QVSRPAMEvents::QVSRPAMEvents(IRPAMTerminalPtr Term)
   {
   m_cref = 0;
   pTerm = Term;
   }

/*
Event handler destructor
*/
QVSRPAMEvents::~QVSRPAMEvents()
   {
   }

/*
Call this routine to connect our event handler up to the
RPAMTerminal object.
*/
HRESULT QVSRPAMEvents::HookEvents(IUnknown *pUnkObject, DWORD Cookie)
   {
   /*
   First ask the terminal for it's IConnectionPointContainer
   interface
   */
   IConnectionPointContainer *pcpc = 0;
   HRESULT hr =
     pUnkObject->QueryInterface(IID_IConnectionPointContainer, (void **)&pcpc);

   if (SUCCEEDED(hr))
      {
      /*
      If that worked then ask the connection point container if
      he supports calling back on an interface that implements
      IQVSRPAMEvents
      */
      IConnectionPoint *pcp = 0;
      hr = pcpc->FindConnectionPoint(__uuidof(_IQVSRPAMEvents), &pcp);
      if (SUCCEEDED(hr))
         {
         /*
         If he says that's OK then pass him a pointer to our implementation
         and he will use it to call back into us.
         */
         hr = pcp->Advise(this, &Cookie);
         pcp->Release();
         }

      pcpc->Release();

      }

   return hr;
   }




/*****************************************************************************/
/* Translate a character to an SA function code                              */
/*****************************************************************************/
long GetFCodeFromKey(int InKey)
   {
   long RetFCode = 0;
   int idx;
   for(idx=0;(KeyMap[idx].Inkey != 0);idx++)
      {
      if (KeyMap[idx].Inkey == InKey)
         {
         RetFCode = KeyMap[idx].FCode;
         break;
         }
      }
   return RetFCode;
   }


/*****************************************************************************/
/* hex logger                                                                */
/*****************************************************************************/
void hexlog(char far *buff,            /* Pointer to data to be logged       */
           int len)                    /* length of data to be logged.       */
   {
   int i,j;                            /* General Loop Control Variables.    */
   struct tmptype                      /* Layout of the Trace message.       */
      {
      char off[6];                     /* Hex address within bufffer.        */
      char hex[16][3];                 /* Hex display locations              */
      char star1[2];                   /* ' *' delimiter of ASCII            */
      char ascii[16];                  /* Actual ASCII string                */
      char star2;                      /* '*' delimiter of ASCII             */
      char crlfnull[3];                /* \r\n\0 end of string.              */
      } tmpbuff;                       /* buffer for layout of data.         */
   static char hexstr[] = "0123456789ABCDEF";
   for (i = 0; i < len; i += 16)       /* Loop for each line of message.     */
      {
      memmove(&tmpbuff,                /* Initialize the buffer.             */
/*
"0000: 00 11 22 33 44 55 66 77 88 99 aa bb cc dd ee ff  *0123456789abcdef*"
*/
"    :                                                  *                *\r\n\x000"
                                              , sizeof(tmpbuff));
      tmpbuff.off[0] = hexstr[(i >> 0x0C) & 0x0F];
                                       /* Fill High order nibble of byte.    */
      tmpbuff.off[1] = hexstr[(i >> 0x08) & 0x0F];
                                       /* Fill Low order nibble of byte.     */
      tmpbuff.off[2] = hexstr[(i >> 0x04) & 0x0F];
                                       /* Fill High order nibble of byte.    */
      tmpbuff.off[3] = hexstr[i & 0x0F];
                                       /* Fill Low order nibble of byte.     */

      for (j = 0;((j < 16) && ((j + i) < len)); j ++)
         {                             /* Loop for each character.           */
         tmpbuff.hex[j][0] = hexstr[(buff[i+j] >> 4)];
                                       /* Fill High order nibble of byte.    */
         tmpbuff.hex[j][1] = hexstr[(buff[i+j] & 0xF)];
                                       /* Fill Low order nibble of byte.     */
         if ((buff[i+j] > ' ') && (buff[i+j] < '~'))
                                       /* If displayable ASCII data          */
            tmpbuff.ascii[j] = buff[i+j];
                                       /* then fill in the ascii byte.       */
         }                             /* End of For 0 < 16                  */
      printf("%s", &tmpbuff);
                                       /* Log the data                       */
      }                                /* End of For 0 < len                 */
   }                                   /* End of HEXLOG();                   */
